package com.liantis.octopusplugin;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import com.nomagic.uml2.ext.magicdraw.auxiliaryconstructs.mdmodels.Model;
import com.nomagic.uml2.ext.magicdraw.auxiliaryconstructs.mdtemplates.TemplateBinding;
import com.nomagic.uml2.ext.magicdraw.auxiliaryconstructs.mdtemplates.TemplateParameterSubstitution;
import com.nomagic.uml2.ext.magicdraw.classes.mdassociationclasses.AssociationClass;
import com.nomagic.uml2.ext.magicdraw.classes.mdinterfaces.Interface;
import com.nomagic.uml2.ext.magicdraw.classes.mdinterfaces.InterfaceRealization;
import com.nomagic.uml2.ext.magicdraw.classes.mdkernel.Association;
import com.nomagic.uml2.ext.magicdraw.classes.mdkernel.Classifier;
import com.nomagic.uml2.ext.magicdraw.classes.mdkernel.DataType;
import com.nomagic.uml2.ext.magicdraw.classes.mdkernel.Element;
import com.nomagic.uml2.ext.magicdraw.classes.mdkernel.Enumeration;
import com.nomagic.uml2.ext.magicdraw.classes.mdkernel.EnumerationLiteral;
import com.nomagic.uml2.ext.magicdraw.classes.mdkernel.Generalization;
import com.nomagic.uml2.ext.magicdraw.classes.mdkernel.MultiplicityElement;
import com.nomagic.uml2.ext.magicdraw.classes.mdkernel.Operation;
import com.nomagic.uml2.ext.magicdraw.classes.mdkernel.Package;
import com.nomagic.uml2.ext.magicdraw.classes.mdkernel.Parameter;
import com.nomagic.uml2.ext.magicdraw.classes.mdkernel.Property;
import com.nomagic.uml2.ext.magicdraw.commonbehaviors.mdbasicbehaviors.Behavior;
import com.nomagic.uml2.ext.magicdraw.commonbehaviors.mdbasicbehaviors.BehavioredClassifier;
import com.nomagic.uml2.ext.magicdraw.deployments.mdnodes.Node;
import com.nomagic.uml2.ext.magicdraw.mdprofiles.Stereotype;
import com.nomagic.uml2.ext.magicdraw.statemachines.mdbehaviorstatemachines.Region;
import com.nomagic.uml2.ext.magicdraw.statemachines.mdbehaviorstatemachines.State;
import com.nomagic.uml2.ext.magicdraw.statemachines.mdbehaviorstatemachines.StateMachine;
import com.nomagic.uml2.omg.classes.kernel.InstanceSpecification;
import com.nomagic.uml2.omg.classes.kernel.LiteralString;
import com.nomagic.uml2.omg.classes.kernel.LiteralBoolean;
import com.nomagic.uml2.omg.classes.kernel.ParameterDirectionKind;
import com.nomagic.uml2.omg.classes.kernel.ParameterDirectionKindEnum;
import com.nomagic.uml2.omg.classes.kernel.Slot;
import com.nomagic.uml2.omg.classes.kernel.Type;
import com.nomagic.uml2.omg.classes.kernel.ValueSpecification;

/**
 * Writes out the current model to octopus uml files
 * 
 * @author costo
 * 
 */
public class OctopusWriter {
    
    private PrintWriter out;
    private Package currentPackage;

    /**
     * writes the uml model to .uml files in the given directoty.
     * 
     * @param directory where .uml files and directories will be written
     * @param model the model to export
     * @throws IOException
     */
    public void writeModel(File directory, Model model) throws IOException {

        java.util.Iterator i = model.getOwnedMember().iterator();
        for (; i.hasNext();) {
            Element element = (Element) i.next();
            if (element instanceof Package) {
                writePackage(directory, (Package) element);
            }
        }
    }

    /**
     * Returns the local or qualified name of the type 'type'.
     * If 'type' is mapped to an OCL type, the OCL typename is
     * returned.
     * 
     * @param type the type where the type name is requested
     * @return the name of the type in OCL
     */
    private String nameForType(Type type) {
        String name = null;

        String oclTypeName = this.getOclType(type);

        if (oclTypeName == null) {
            if (type.getOwningPackage() == currentPackage) {
                name = type.getName();
            } else {
                name = type.getQualifiedName();
            }

            return getOclName(name);
        } else {
            return oclTypeName;
        }
    }

    /**
     * Returns the OCL type name of the passed type.
     * 
     * @param type the type to convert to some ocl type
     * @return the name of the ocl type or null iff type is not mapped to an ocl type
     */
    private String getOclType(Type type) {

        String oclType = getTaggedValue((Element) type, "oclType");

        if (oclType == null && type instanceof DataType) {
            DataType dataType = (DataType) type;
            // check if this is a template instance
            if (dataType.getTemplateBinding().size() > 0) {
                TemplateBinding binding = (TemplateBinding) dataType.getTemplateBinding().iterator().next();
                Type template = (Type) binding.getSignature().getTemplate();
                oclType = getOclType(template);
                if (oclType != null) {
                    if (oclType.equals("Set") || oclType.equals("Sequence") || oclType.equals("Bag")
                            || oclType.equals("OrderedSet")) {
                        if (binding.getParameterSubstitution().size() > 0) {
                            TemplateParameterSubstitution substitution = (TemplateParameterSubstitution) binding
                                    .getParameterSubstitution().iterator().next();
                            Collection actual = substitution.getActual();
                            if (actual.size() > 0) {
                                Object firstActual = actual.iterator().next();
                                if (firstActual instanceof Type) {
                                    oclType = oclType + "(" + this.nameForType((Type) firstActual) + ")";
                                }
                            }
                        }
                    }
                }
            }
        }
        return oclType;
    }

    /**
     * Returns the OCL name of the UML element name
     * 
     * @param umlName the name of the element as defined in the uml model
     * @return the name of the element as used in OCL
     */
    private String getOclName(String umlName) {

        if (umlName == null || umlName.length() == 0) {
            return umlName;
        }

        // append _ if first letter is digit
        // replace non letters with _
        char firstLetter = umlName.charAt(0);
        StringBuffer buf = new StringBuffer(umlName.length());
        if (!Character.isLetter(firstLetter)) {
            buf.append("_");
            if (Character.isDigit(firstLetter)) {
                buf.append(firstLetter);
            }
        } else {
            buf.append(firstLetter);
        }
        for (int i = 1, j = umlName.length(); i < j; i++) {
            char ch = umlName.charAt(i);
            if (Character.isLetter(ch) || Character.isDigit(ch) || ch == ':') {
                buf.append(ch);
            } else {
                buf.append("_");
            }
        }

        return buf.toString();
    }

    /**
     * Returns the string value of the named tagged value
     * @param c the element to query the value
     * @param tagName the name of the tagged value
     * @return the tagged value's value or null
     */
    private String getTaggedValue(Element c, String tagName) {

        String result = null;

        InstanceSpecification stereotypeInstance = c.getAppliedStereotypeInstance();
        if (stereotypeInstance != null) {
            Iterator slots = stereotypeInstance.getSlot().iterator();
            while (slots.hasNext()) {
                Slot slot = (Slot) slots.next();
                if (slot.getDefiningFeature().getName().equals(tagName)) {
                    if (slot.getValue().size() > 0) {
                        ValueSpecification value = (ValueSpecification)slot.getValue().get(0);
                        if(value instanceof LiteralString) {
                            result = ((LiteralString)value).getValue();
                        }
                        else if(value instanceof LiteralBoolean) {
                            result = ((LiteralBoolean)value).isValue()+"";                            
                        }
                        break;
                    }
                }
            }
        }

        return result;
    }

    /**
     * Writes the content of the given package
     * @param path
     * @param aPackage
     * @throws IOException
     */
    private void writePackage(File path, Package aPackage) throws IOException {

        if (aPackage.getName().startsWith("UML Standard Profile")) {
            return;
        }

        // collect elements to export
        List classifiers = new ArrayList();
        List datatypes = new ArrayList();
        List associations = new ArrayList();

        for (java.util.Iterator i = aPackage.getOwnedMember().iterator(); i.hasNext();) {
            Element element = (Element) i.next();
            if ((element instanceof AssociationClass) || (element instanceof Enumeration)
                    || (element instanceof Interface)) {

                classifiers.add(element);

            } else if (element instanceof com.nomagic.uml2.ext.magicdraw.classes.mdkernel.Class) {
                if (!(element instanceof Behavior || element instanceof Node || element instanceof Stereotype)) {

                    classifiers.add(element);

                }
            } else if (element instanceof DataType) {

                String oclType = this.getOclType((DataType)element);
                // write datatype only if its not(!) mapped to OCL type
                if(oclType == null) {
                    datatypes.add(element);
                }

            } else if (element instanceof Association) {
                associations.add(element);
            }
        }

        // create directory
        File newPath = new File(path, getOclName(aPackage.getName()));
        newPath.mkdir();
        path = newPath;
        currentPackage = aPackage;

        // write all classifiers each in its own file
        try {
            for (int i = 0, j = classifiers.size(); i < j; i++) {

                Classifier classifier = (Classifier)classifiers.get(i);
                File file = new File(path, getOclName(classifier.getName()) + ".uml");
                out = new PrintWriter(file);

                out.print("<package> ");
                out.println(getOclName(aPackage.getName()));

                if (classifier instanceof AssociationClass) {
                    writeAssociationClass((AssociationClass) classifier);
                } else if (classifier instanceof com.nomagic.uml2.ext.magicdraw.classes.mdkernel.Class) {
                    writeClass((com.nomagic.uml2.ext.magicdraw.classes.mdkernel.Class) classifier);
                } else if (classifier instanceof Enumeration) {
                    writeEnumeration((Enumeration) classifier);
                } else if (classifier instanceof Interface) {
                    writeInterface((Interface) classifier);
                }

                out.println("<endpackage>");
                out.close();
                out = null;
            }

            // write all datatypes into file datatypes.uml
            if (datatypes.size() > 0) {

                File file = new File(path, "datatypes.uml");
                out = new PrintWriter(file);

                out.print("<package> ");
                out.println(getOclName(aPackage.getName()));

                for (int i = 0, j = datatypes.size(); i < j; i++) {

                    DataType dataType = (DataType) datatypes.get(i);
                    writeDatatype(dataType);

                }

                out.println("<endpackage>");
                out.close();
                out = null;
            }

            // write all associations in one file associations.uml
            if (associations.size() > 0) {

                File file = new File(path, "associations.uml");
                out = new PrintWriter(file);

                out.print("<package> ");
                out.println(getOclName(aPackage.getName()));
                out.println("<associations>");
                for (int i = 0, j = associations.size(); i < j; i++) {

                    Association association = (Association) associations.get(i);
                    writeAssociation(association);

                }

                out.println("<endpackage>");
                out.close();
                out = null;
            }
        } finally {
            if (out != null) {
                out.close();
                out = null;
            }
        }
        
        // write all subpackages
        for (Iterator i = aPackage.getOwnedMember().iterator(); i.hasNext();) {
            Element element = (Element) i.next();
            if (element instanceof Package) {
                Package p = (Package) element;
                writePackage(path, p);
            }
        }

    }

    /**
     * Writes the abstract property of the given classifier
     * @param c the classifier
     */
    private void writeClassifier(Classifier c) {

        if (c.isAbstract()) {
            out.print("<abstract>");
        }

    }

    /**
     * Writes out the given class
     * @param c
     */
    private void writeClass(com.nomagic.uml2.ext.magicdraw.classes.mdkernel.Class c) {
        writeClassifier(c);
        out.print("<class> ");
        out.println(getOclName(c.getName()));
        writeGeneralizationsForClassifier(c);
        writeInterfaceRealizationsForClassifier(c);
        writeAttributes(c);
        writeOperations(c);
        writeStates(c);
        out.println("<endclass>");
    }

    /**
     * Writes out the given datatype or primitive
     * @param dataType
     */
    private void writeDatatype(DataType dataType) {

        writeClassifier(dataType);
        out.print("<datatype> ");
        out.println(getOclName(dataType.getName()));
        writeGeneralizationsForClassifier(dataType);
        writeAttributes(dataType);
        writeOperations(dataType);
        out.println("<enddatatype>");
    }

    /**
     * Writes out the given interface
     * @param interf
     */
    private void writeInterface(Interface interf) {
        writeClassifier(interf);
        out.print("<interface> ");
        out.println(getOclName(interf.getName()));
        writeGeneralizationsForClassifier(interf);
        writeAttributes(interf);
        writeOperations(interf);
        out.println("<endinterface>");
    }

    /**
     * Writes out the given enumeration
     * @param enumeration
     */
    private void writeEnumeration(Enumeration enumeration) {
        writeClassifier(enumeration);
        out.print("<enumeration> ");
        out.print(getOclName(enumeration.getName()));
        out.println(" <values> ");
        Iterator iter = enumeration.getOwnedLiteral().iterator();
        while (iter.hasNext()) {
            EnumerationLiteral lit = (EnumerationLiteral) iter.next();
            out.print(getOclName(lit.getName()));
            out.println(";");
        }

        out.println("<endenumeration>");

    }

    /**
     * Writes out the given association class
     * @param assClass
     */
    private void writeAssociationClass(AssociationClass assClass) {
        writeClassifier(assClass);
        out.print("<associationclass> ");
        out.println(getOclName(assClass.getName()));
        Property end1 = (Property) assClass.getMemberEnd().get(0);
        Property end2 = (Property) assClass.getMemberEnd().get(1);
        writeAssociationEnd(end1);
        writeNavigability(end1, end2);
        writeAssociationEnd(end2);
        out.println("");
        writeAttributes(assClass);
        writeOperations(assClass);
        writeStates(assClass);
        out.println("<endassociationclass>");
    }

    /**
     * Writes out the given association
     * @param association
     */
    private void writeAssociation(Association association) {
        if (association.isDerived()) {
            out.print("/ ");
        }
        String name = getOclName(association.getName());
        if (name != null && name.length() > 0) {
            out.print(name + " : ");
        }
        Property end1 = (Property) association.getMemberEnd().get(0);
        Property end2 = (Property) association.getMemberEnd().get(1);
        writeAssociationEnd(end1);
        writeNavigability(end1, end2);
        writeAssociationEnd(end2);
        out.println(";");
    }

    private void writeAssociationEnd(Property end) {
        if (end.isStatic()) {
            out.print("$");
        }
        String name = getOclName(end.getName());
        String roleName = (name != null && name.length() > 0) ? name : "<noName>";
        out.print(nameForType((Classifier) end.getType()) + "." + roleName);
        writeMultiplicity(end);
        if (end.isOrdered()) {
            out.print("<ordered>");
        }
        if (!end.isUnique()) {
            out.print("<notUnique>");
        }
        out.print(" ");

    }

    private void writeNavigability(Property end1, Property end2) {
        if (end1.isNavigable() && end2.isNavigable()) {
            out.print("<-> ");
        } else if (!end1.isNavigable() && end2.isNavigable()) {
            out.print("-> ");
        } else if (end1.isNavigable() && !end2.isNavigable()) {
            out.print("<- ");
        } else {
            out.print("- ");
        }
    }

    private boolean writeMultiplicity(MultiplicityElement mult) {
        String upper = null;
        boolean many = false;
        if (mult.getUpper() == -1) {
            upper = "*";
            many = true;
        } else if (mult.getUpper() == 0) {
            upper = "1";
        } else {
            upper = "" + mult.getUpper();
            many = (mult.getUpper() > 1);
        }
        out.print("[" + mult.getLower() + ".." + upper + "]");

        return many;
    }

    private void writeAttributes(Classifier c) {
        boolean first = true;
        Iterator iter = c.getAttribute().iterator();
        while (iter.hasNext()) {
            Object nextObject = iter.next();
            if (nextObject instanceof Property) {
                Property prop = (Property) nextObject;
                if (prop.getAssociation() == null) {
                    if (first) {
                        out.println("<attributes>");
                        first = false;
                    }
                    if (prop.isStatic()) {
                        out.print("$ ");
                    }
                    if (prop.isDerived() || prop.isDerivedUnion()) {
                        out.print("/ ");
                    }
                    out.print(getOclName(prop.getName()) + " ");
                    boolean isMany = writeMultiplicity(prop);
                    out.print(" : ");
                    if (isMany) {
                        if (prop.isUnique() && prop.isOrdered()) {
                            out.print("OrderedSet(");
                        } else if (!prop.isUnique() && prop.isOrdered()) {
                            out.print("Sequence(");
                        } else if (prop.isUnique() && !prop.isOrdered()) {
                            out.print("Set(");
                        } else {
                            out.print("Bag(");
                        }
                    }
                    out.print(prop.getType() != null ? nameForType(prop.getType()) : "unknown");
                    if (isMany)
                        out.print(")");
                    out.println(" ;");
                }
            }
        }
    }

    private void writeOperations(Classifier c) {
        Iterator features = c.getFeature().iterator();
        boolean first = true;
        while (features.hasNext()) {
            Object nextObject = features.next();
            if (nextObject instanceof Operation) {
                if (first) {
                    out.println("<operations>");
                    first = false;
                }
                Operation operation = (Operation) nextObject;
                if (operation.isStatic()) {
                    out.print("$ ");
                }
                if (operation.isAbstract()) {
                    out.println("<abstract> ");
                }
                String operationName = operation.getName();
                String infix = getTaggedValue(operation,"oclInfix");
                boolean isInfix = (operationName.equals("+") || operationName.equals("-") || operationName.equals("/")
                        || operationName.equals("<") || operationName.equals(">") || operationName.equals("<=")
                        || operationName.equals(">=") || operationName.equals("=") || operationName.equals("<>")
                        || operationName.equals("=") || operationName.equals("<>") || operationName.equals("*")) ||
                    (infix != null && infix.equals("true"));
                
                if(isInfix) {
                    out.print("<infix>"+operationName);
                }
                else {
                    out.print(getOclName(operationName));
                }
                out.print("(");

                Type returnType = null;
                Parameter previousParameter = null;

                Iterator params = operation.getOwnedParameter().iterator();
                while (params.hasNext()) {
                    Parameter parameter = (Parameter) params.next();
                    ParameterDirectionKind direction = parameter.getDirection();
                    if (direction != null) {
                        if (direction == ParameterDirectionKindEnum.RETURN) {
                            returnType = parameter.getType();
                            continue;
                        }
                        if (previousParameter != null) {
                            out.print(", ");
                        }
                        previousParameter = parameter;
                        
                        if (direction == ParameterDirectionKindEnum.IN) {
                            out.print("<in>");
                        }
                        if (direction == ParameterDirectionKindEnum.OUT) {
                            out.print("<out>");
                        }
                        if (direction == ParameterDirectionKindEnum.INOUT) {
                            out.print("<inout>");
                        }
                    }
                    out.print(getOclName(parameter.getName()));
                    out.print(":");
                    out.print(parameter.getType() != null ? nameForType(parameter.getType()) : "unknown");
                }
                out.print(")");
                if (returnType != null) {
                    out.print(" : " + nameForType(returnType));
                }
                out.println(";");
            }

        }
    }

    /**
     * If the classifier behaviour of the given classifier is a state machine,
     * writes out all states of this state machine.
     * @param c the behaviored classifier
     */
    private void writeStates(BehavioredClassifier c) {
        if (c.getClassifierBehavior() != null && c.getClassifierBehavior() instanceof StateMachine) {
            out.println("<states>");
            StateMachine stateMachine = (StateMachine) c.getClassifierBehavior();
            Iterator regions = stateMachine.getRegion().iterator();
            while (regions.hasNext()) {
                Region region = (Region) regions.next();
                writeStates("", region);
            }
        }
    }

    private void writeStates(String prefix, Region region) {
        Iterator vertex = region.getSubvertex().iterator();
        while (vertex.hasNext()) {
            Object object = vertex.next();
            if (object instanceof State) {
                State state = (State) object;
                String name = prefix + getOclName(state.getName());
                out.println(name + ";");

                if (state.isComposite()) {
                    Iterator regions = state.getRegion().iterator();
                    while (regions.hasNext()) {
                        Region subregion = (Region) regions.next();
                        writeStates(name + "::", subregion);
                    }
                }
            }
        }
    }

    private void writeGeneralizationsForClassifier(com.nomagic.uml2.ext.magicdraw.classes.mdkernel.Classifier c) {
        Collection generalizations = c.getGeneralization();
        if (generalizations.size() > 0) {
            out.print("<specializes> ");
            Iterator i = generalizations.iterator();
            while (i.hasNext()) {
                Generalization g = (Generalization) i.next();
                out.print(nameForType(g.getGeneral()));
                if (i.hasNext()) {
                    out.print(",");
                } else {
                    out.println("");
                }
            }
        }
    }

    private void writeInterfaceRealizationsForClassifier(BehavioredClassifier c) {
        Collection ir = c.getInterfaceRealization();
        if (ir.size() > 0) {
            Iterator i = ir.iterator();
            out.print("<implements> ");
            while (i.hasNext()) {
                InterfaceRealization interfaceR = (InterfaceRealization) i.next();
                Collection target = interfaceR.getTarget();
                if (target.size() == 1) {
                    out.print(nameForType((Type) target.iterator().next()));
                    if (i.hasNext()) {
                        out.print(",");
                    } else {
                        out.println("");
                    }
                }
            }
        }
    }

}
